import logging

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from tinkerforge.bricklet_humidity_v2 import BrickletHumidityV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase, print_exception_decorator
from tinkerforge2mqtt.utilities.dew_point import calculate_dew_point


logger = logging.getLogger(__name__)


@register_map_class()
class BrickletHumidityV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/HumidityV2_Bricklet_Python.html

    device_identifier = BrickletHumidityV2.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletHumidityV2, **kwargs):
        self.device: BrickletHumidityV2 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.humidity = Sensor(
            device=self.mqtt_device,
            name='Humidity',
            uid='humidity',
            device_class='humidity',
            state_class='measurement',
            unit_of_measurement='%',
            suggested_display_precision=2,
        )
        self.temperature = Sensor(
            device=self.mqtt_device,
            name='Temperature',
            uid='temperature',
            device_class='temperature',
            state_class='measurement',
            unit_of_measurement='°C',
            suggested_display_precision=2,
        )
        self.dew_point = Sensor(
            device=self.mqtt_device,
            name='Dew Point Temperature',
            uid='dew_point',
            device_class='temperature',
            state_class='measurement',
            unit_of_measurement='°C',
            suggested_display_precision=2,
        )

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.set_temperature_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_TEMPERATURE, self.callback_temperature)

        self.device.set_humidity_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_HUMIDITY, self.callback_humidity)

    @print_exception_decorator
    def callback_temperature(self, value):
        temperature = value / 100
        logger.info(f'Temperature callback: {temperature}°C (UID: {self.device.uid_string})')
        self.temperature.set_state(temperature)
        self.temperature.publish(self.mqtt_client)
        self.set_and_publish_dew_point()

    @print_exception_decorator
    def callback_humidity(self, value):
        humidity = value / 100
        logger.info(f'Humidity callback: {humidity}% (UID: {self.device.uid_string})')
        self.humidity.set_state(humidity)
        self.humidity.publish(self.mqtt_client)
        self.set_and_publish_dew_point()

    def set_and_publish_dew_point(self):
        if self.temperature.value and self.humidity.value:
            dew_point = calculate_dew_point(
                temperature=self.temperature.value,
                humidity=self.humidity.value,
            )
            logger.info(f'Dew Point: {dew_point}°C (UID: {self.device.uid_string})')
            self.dew_point.set_state(dew_point)
            self.dew_point.publish(self.mqtt_client)



import logging

from ha_services.mqtt4homeassistant.components.binary_sensor import BinarySensor
from tinkerforge.bricklet_motion_detector_v2 import BrickletMotionDetectorV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickletMotionDetectorV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/MotionDetectorV2_Bricklet_Python.html

    device_identifier = BrickletMotionDetectorV2.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletMotionDetectorV2, **kwargs):
        self.device: BrickletMotionDetectorV2 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.motion_detected = BinarySensor(
            device=self.mqtt_device,
            name='Motion Detected',
            uid='motion_detected',
            device_class='motion',
        )
        logger.info(f'Creating: {self.motion_detected}')

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.register_callback(self.device.CALLBACK_MOTION_DETECTED, self.callback_motion_detected)
        self.device.register_callback(self.device.CALLBACK_DETECTION_CYCLE_ENDED, self.callback_detection_cycle_ended)

    @print_exception_decorator
    def callback_motion_detected(self):
        logger.debug(f'Motion detected (UID: {self.device.uid_string})')
        self.motion_detected.set_state(self.motion_detected.ON)
        self.motion_detected.publish(self.mqtt_client)

    @print_exception_decorator
    def callback_detection_cycle_ended(self):
        logger.debug(f'Detection Cycle Ended (UID: {self.device.uid_string})')
        self.motion_detected.set_state(self.motion_detected.OFF)
        self.motion_detected.publish(self.mqtt_client)



from functools import wraps

from rich.console import Console


def print_exception_decorator(func):

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            console = Console()
            console.print_exception(show_locals=True)
            raise SystemExit from err

    return func_wrapper



import logging

from cli_base.autodiscover import import_all_files
from tinkerforge.device_factory import DEVICE_CLASSES

from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase


logger = logging.getLogger(__name__)


class MapRegistry:
    def __init__(self):
        self._registry = {}

    def add_map_class(self, MapClass):
        logger.debug(f'Add map class: {MapClass}')
        assert issubclass(MapClass, DeviceMapBase), f'Class {MapClass} must be subclass of {DeviceMapBase}'
        device_identifier = MapClass.device_identifier
        assert device_identifier in DEVICE_CLASSES, f'Unknown device identifier: {device_identifier} from {MapClass}'
        self._registry[device_identifier] = MapClass

    def get_map_class(self, device_identifier) -> type[DeviceMapBase] | None:
        return self._registry.get(device_identifier)


map_registry = MapRegistry()


def register_map_class():
    """
    Decorator to add a new map class to the map registry.
    """

    def wrapper(MapClass):
        map_registry.add_map_class(MapClass)
        return MapClass

    return wrapper


# Register all map classes, just by import all files in this package:
import_all_files(package=__package__, init_file=__file__)



import math


SV_PRESSURE = 6.112  # saturation vapor pressure at 0°C


def calculate_dew_point(
    temperature: int | float,  # Temperature in degrees Celsius
    humidity: int | float,  # Relative Humidity in %
) -> float:  # Returns: Dew point temperature in degrees Celsius
    """
    Calculate the dew point temperature given the temperature and relative humidity
    using the Magnus formula.

    Examples:
    >>> round(calculate_dew_point(temperature=18, humidity=40), 1)
    4.2
    >>> round(calculate_dew_point(temperature=22, humidity=70), 1)
    16.3
    >>> round(calculate_dew_point(temperature=30, humidity=50), 1)
    18.4
    >>> round(calculate_dew_point(temperature=-5, humidity=80), 1)
    -7.6
    """
    # Constants for the Magnus formula
    if temperature > 0:
        # above freezing
        a = 17.62  # °C
        b = 243.12  # °C
    else:
        # below freezing
        a = 22.46  # °C
        b = 272.62  # °C

    es = SV_PRESSURE * math.exp((a * temperature) / (b + temperature))  # saturation vapor pressure
    e = es * (humidity / 100.0)  # actual vapor pressure

    # Solve for the dew point temperature:
    dew_point_temperature = (b * math.log(e / SV_PRESSURE)) / (a - math.log(e / SV_PRESSURE))

    return dew_point_temperature



import logging

from ha_services.mqtt4homeassistant.components.switch import Switch
from paho.mqtt.client import Client
from tinkerforge.bricklet_industrial_dual_relay import BrickletIndustrialDualRelay

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickletIndustrialDualRelayMapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/IndustrialDualRelay_Bricklet_Python.html

    device_identifier = BrickletIndustrialDualRelay.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletIndustrialDualRelay, **kwargs):
        self.device: BrickletIndustrialDualRelay = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.relay0_switch = Switch(
            device=self.mqtt_device,
            name='Relay 0',
            uid='relay0',
            callback=self.relay0_callback,
        )
        logger.info(f'Creating: {self.relay0_switch}')

        self.relay1_switch = Switch(
            device=self.mqtt_device,
            name='Relay 1',
            uid='relay1',
            callback=self.relay1_callback,
        )
        logger.info(f'Creating: {self.relay1_switch}')

    @print_exception_decorator
    def setup_callbacks(self):
        pass  # TODO

    @print_exception_decorator
    def poll(self):
        super().poll()

        relay0value, relay1value = self.device.get_value()

        if relay0value:
            self.relay0_switch.set_state(self.relay0_switch.ON)
        else:
            self.relay0_switch.set_state(self.relay0_switch.OFF)
        self.relay0_switch.publish(self.mqtt_client)

        if relay1value:
            self.relay1_switch.set_state(self.relay1_switch.ON)
        else:
            self.relay1_switch.set_state(self.relay1_switch.OFF)
        self.relay1_switch.publish(self.mqtt_client)

    @print_exception_decorator
    def relay0_callback(self, *, client: Client, component: Switch, old_state: str, new_state: str):
        logger.error(f'{component.name} relay 0 state changed: {old_state!r} -> {new_state!r}')
        turn_relay_on = new_state == self.relay0_switch.ON
        self.device.set_selected_value(channel=0, value=turn_relay_on)

        self.poll()

    @print_exception_decorator
    def relay1_callback(self, *, client: Client, component: Switch, old_state: str, new_state: str):
        logger.error(f'{component.name} relay 1 state changed: {old_state!r} -> {new_state!r}')
        turn_relay_on = new_state == self.relay1_switch.ON
        self.device.set_selected_value(channel=1, value=turn_relay_on)

        self.poll()



import logging
import sys
import time
from pathlib import Path

import rich_click as click
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE, setup_logging
from rich import print  # noqa
from tinkerforge.device_factory import get_device_class
from tinkerforge.ip_connection import Error, IPConnection

from tinkerforge2mqtt.cli_app import cli
from tinkerforge2mqtt.cli_app.settings import get_user_settings
from tinkerforge2mqtt.device_map_utils.generics import iter_interest_functions
from tinkerforge2mqtt.device_registry.devices_handler import DevicesHandler
from tinkerforge2mqtt.user_settings import UserSettings


logger = logging.getLogger(__name__)


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE | {'default': 2})
def local_hat_info(verbosity: int):
    """
    Just print information about from `/proc/device-tree/hat/` files.
    """
    setup_logging(verbosity=verbosity)
    base_path = Path('/proc/device-tree/hat/')
    if not base_path.is_dir():
        print(f'ERROR: Path not found: {base_path}')
        sys.exit(-1)

    for file_path in base_path.glob('*'):
        try:
            content = file_path.read_text()
        except Exception:
            logger.exception(f'Can not read file {file_path}')
        else:
            print(f'{file_path.name}: {content}')


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE | {'default': 2})
def discover(verbosity: int):
    """
    Discover Victron devices with Instant Readout
    """
    setup_logging(verbosity=verbosity)
    user_settings: UserSettings = get_user_settings(verbosity=verbosity)

    # https://github.com/Tinkerforge/generators/tree/master/mqtt

    # https://www.tinkerforge.com/en/doc/Software/IPConnection_Python.html
    ipcon = IPConnection()
    connect_kwargs = dict(
        host=user_settings.host,
        port=user_settings.port,
    )
    print(f'Connecting to {connect_kwargs}')
    ipcon.connect(**connect_kwargs)

    def enumerate_handler(
        uid,
        connected_uid,
        position,
        hardware_version,
        firmware_version,
        device_identifier,
        enumeration_type,
    ):
        DeviceClass = get_device_class(device_identifier)
        print('_' * 80)
        print(f'{DeviceClass.DEVICE_DISPLAY_NAME} ({DeviceClass.__name__})')

        if enumeration_type == IPConnection.ENUMERATION_TYPE_DISCONNECTED:
            print('Disconnected!')
            return

        device = DeviceClass(uid=uid, ipcon=ipcon)

        for func in iter_interest_functions(device):
            name = func.__name__
            print(f'{name}():', end=' ')
            try:
                value = func()
            except Error as e:
                print(f'ERROR: {e}')
            else:
                print(value)

    ipcon.register_callback(IPConnection.CALLBACK_ENUMERATE, enumerate_handler)
    ipcon.enumerate()
    time.sleep(5)
    ipcon.disconnect()


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE | {'default': 2})
def discover_map(verbosity: int):
    """
    Discover Victron devices with Instant Readout
    """
    setup_logging(verbosity=verbosity)
    user_settings: UserSettings = get_user_settings(verbosity=verbosity)

    # https://github.com/Tinkerforge/generators/tree/master/mqtt

    # https://www.tinkerforge.com/en/doc/Software/IPConnection_Python.html
    ipcon = IPConnection()
    connect_kwargs = dict(
        host=user_settings.host,
        port=user_settings.port,
    )
    print(f'Connecting to {connect_kwargs}')
    ipcon.connect(**connect_kwargs)

    devices_handler = DevicesHandler(ipcon)

    ipcon.register_callback(IPConnection.CALLBACK_ENUMERATE, devices_handler)

    print('Aborting with Ctrl-C !')
    while True:
        try:
            ipcon.enumerate()
            time.sleep(5)
        except KeyboardInterrupt:
            logger.info('Keyboard interrupt')
            ipcon.disconnect()
            break



import logging
import socket

from ha_services.mqtt4homeassistant.device import MainMqttDevice
from paho.mqtt.client import Client
from tinkerforge.device_factory import get_device_class
from tinkerforge.ip_connection import Device, IPConnection

import tinkerforge2mqtt
from tinkerforge2mqtt.device_map import map_registry
from tinkerforge2mqtt.user_settings import UserSettings


logger = logging.getLogger(__name__)


class DevicesHandler:
    def __init__(
        self,
        ipcon: IPConnection,
        *,
        mqtt_client: Client,
        user_settings: UserSettings,
    ):
        self.ipcon = ipcon
        self.mqtt_client = mqtt_client
        self.user_settings = user_settings

        self.main_mqtt_device = MainMqttDevice(
            name=f'tinkerforge2mqtt@{socket.gethostname()}',
            uid=user_settings.mqtt.main_uid,
            manufacturer='tinkerforge2mqtt',
            sw_version=tinkerforge2mqtt.__version__,
        )

        self.map_instances = {}

    def __call__(
        self,
        uid,
        connected_uid,
        position,
        hardware_version,
        firmware_version,
        device_identifier,
        enumeration_type,
    ):
        if enumeration_type == IPConnection.ENUMERATION_TYPE_DISCONNECTED:
            logger.warning(f'Disconnected: {uid=}')
            return

        if map_instance := self.map_instances.get(uid):
            logger.debug(f'Already initialized: {uid=} {device_identifier=} {map_instance=}')
        else:
            logger.info(f'New device: {uid=} {device_identifier=}')

            TinkerforgeDeviceClass = get_device_class(device_identifier)
            name = f'{TinkerforgeDeviceClass.DEVICE_DISPLAY_NAME} ({TinkerforgeDeviceClass.__name__})'
            logger.info(name)

            MapClass = map_registry.get_map_class(device_identifier)
            if not MapClass:
                logger.error(f'No mapper found for {TinkerforgeDeviceClass.__name__} ({device_identifier=})')
                return

            device: Device = TinkerforgeDeviceClass(
                uid=uid,
                ipcon=self.ipcon,
            )

            map_instance = MapClass(
                main_mqtt_device=self.main_mqtt_device,
                device=device,
                mqtt_client=self.mqtt_client,
                user_settings=self.user_settings,
            )
            self.map_instances[uid] = map_instance

        map_instance.poll()

    def connected_handler(self, *args, **kwargs):
        print('Connected!', args, kwargs, self.ipcon.devices)

    def disconnected_handler(self, *args, **kwargs):
        print('Disconnected!', args, kwargs)



import logging

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from tinkerforge.bricklet_voltage_current_v2 import BrickletVoltageCurrentV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickletVoltageCurrentV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/VoltageCurrentV2_Bricklet_Python.html

    device_identifier = BrickletVoltageCurrentV2.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletVoltageCurrentV2, **kwargs):
        self.device: BrickletVoltageCurrentV2 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.current = Sensor(
            device=self.mqtt_device,
            name='Current',
            uid='current',
            device_class='current',
            state_class='measurement',
            unit_of_measurement='A',
            suggested_display_precision=3,
        )
        self.voltage = Sensor(
            device=self.mqtt_device,
            name='Voltage',
            uid='voltage',
            device_class='voltage',
            state_class='measurement',
            unit_of_measurement='V',
            suggested_display_precision=3,
        )
        self.power = Sensor(
            device=self.mqtt_device,
            name='Power',
            uid='power',
            device_class='power',
            state_class='measurement',
            unit_of_measurement='W',
            suggested_display_precision=3,
        )

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.set_voltage_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_VOLTAGE, self.callback_voltage)

        self.device.set_current_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_CURRENT, self.callback_current)

        self.device.set_power_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_POWER, self.callback_power)

    @print_exception_decorator
    def callback_voltage(self, value):
        logger.info(f'Voltage callback: {value}mV (UID: {self.device.uid_string})')
        voltage = value / 1000.0
        self.voltage.set_state(voltage)
        self.voltage.publish(self.mqtt_client)

    @print_exception_decorator
    def callback_current(self, value):
        logger.info(f'Current callback: {value}mA (UID: {self.device.uid_string})')
        current = value / 1000.0
        self.current.set_state(current)
        self.current.publish(self.mqtt_client)

    @print_exception_decorator
    def callback_power(self, value):
        logger.info(f'Current callback: {value}mW (UID: {self.device.uid_string})')
        power = value / 1000.0
        self.power.set_state(power)
        self.power.publish(self.mqtt_client)



import logging

from ha_services.mqtt4homeassistant.components import BaseComponent
from ha_services.mqtt4homeassistant.components.select import Select
from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify
from paho.mqtt.client import Client
from tinkerforge.bricklet_analog_in_v3 import BrickletAnalogInV3

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator


logger = logging.getLogger(__name__)


INTERVAL_DURATION = 0.0175  # 17,5µs
OVERSAMPLING_MAP = {
    0: 32,  # 32x17,5µs = 0.56ms
    1: 64,
    2: 128,
    3: 256,
    4: 512,
    5: 1024,
    6: 2048,
    7: 4096,  # 4096x17,5µs = 71.68ms **the default**
    8: 8192,
    9: 16384,  # 16384x17,5µs = 286.72ms
}


def human_readable_oversampling(oversampling: int):
    """
    >>> human_readable_oversampling(0)
    '32x (0.56ms)'
    >>> human_readable_oversampling(7)
    '4096x (71.68ms)'
    >>> human_readable_oversampling(9)
    '16384x (286.72ms)'
    """
    assert oversampling in OVERSAMPLING_MAP, f'Invalid: {oversampling=}'
    factor = OVERSAMPLING_MAP[oversampling]
    duration = factor * INTERVAL_DURATION
    return f'{factor}x ({duration:.2f}ms)'


OVERSAMPLE2OPTION = {
    oversampling: slugify(human_readable_oversampling(oversampling)) for oversampling in OVERSAMPLING_MAP.keys()
}
OPTION2OVERSAMPLE = {v: k for k, v in OVERSAMPLE2OPTION.items()}


@register_map_class()
class BrickletAnalogInV3Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/AnalogInV3_Bricklet_Python.html

    device_identifier = BrickletAnalogInV3.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletAnalogInV3, **kwargs):
        self.device: BrickletAnalogInV3 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.voltage = Sensor(
            device=self.mqtt_device,
            name='Voltage',
            uid='voltage',
            device_class='voltage',
            state_class='measurement',
            unit_of_measurement='V',
            suggested_display_precision=3,
        )

        current_oversampling = self.device.get_oversampling()

        self.oversampling = Select(
            device=self.mqtt_device,
            name='Oversampling',
            uid='oversampling',
            callback=self.oversampling_callback,
            options=tuple(OPTION2OVERSAMPLE.keys()),
            default_option=OVERSAMPLE2OPTION[current_oversampling],
        )

    @print_exception_decorator
    def oversampling_callback(self, *, client: Client, component: BaseComponent, old_state: str, new_state: str):
        logger.info(f'{component.name} state changed: {old_state!r} -> {new_state!r}')
        self.device.set_oversampling(new_state)
        self.poll_oversampling()

    @print_exception_decorator
    def poll_oversampling(self):
        current_oversampling = self.device.get_oversampling()
        option = OVERSAMPLE2OPTION[current_oversampling]
        self.oversampling.set_state(option)
        logger.info(f'{current_oversampling=} {option=} (UID: {self.device.uid_string})')
        self.oversampling.publish(self.mqtt_client)

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.set_voltage_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=0,
            max=0,
        )
        self.device.register_callback(self.device.CALLBACK_VOLTAGE, self.callback_voltage)

    @print_exception_decorator
    def callback_voltage(self, value):
        logger.info(f'Voltage callback: {value}mV (UID: {self.device.uid_string})')
        voltage = value / 1000.0
        self.voltage.set_state(voltage)
        self.voltage.publish(self.mqtt_client)

        # Poll the oversampling value and expose it as MQTT "select"
        self.poll_oversampling()



import inspect
from collections.abc import Callable, Generator

from tinkerforge.ip_connection import Device


def iter_interest_functions(device: Device) -> Generator[Callable, None, None]:
    for name, func in inspect.getmembers(device, inspect.ismethod):
        if '_callback_' in name:
            continue

        if not name.startswith(('get_', 'read_')):
            continue

        if name in ('get_bootloader_mode', 'read_uid'):
            continue

        spec = inspect.getfullargspec(func)
        if spec.args == ['self']:
            yield func



"""
    CLI for usage
"""

import logging
import sys

import rich_click as click
from rich import print  # noqa
from rich_click import RichGroup

from cli_base.autodiscover import import_all_files
from cli_base.cli_tools.version_info import print_version

import tinkerforge2mqtt
from tinkerforge2mqtt import constants


logger = logging.getLogger(__name__)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


# Register all click commands, just by import all files in this package:
import_all_files(package=__package__, init_file=__file__)


@cli.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


def main():
    print_version(tinkerforge2mqtt)

    # Execute Click CLI:
    cli.name = './cli.py'
    cli()



import dataclasses
import sys

from cli_base.systemd.data_classes import BaseSystemdServiceInfo, BaseSystemdServiceTemplateContext
from ha_services.mqtt4homeassistant.data_classes import MqttSettings


@dataclasses.dataclass
class SystemdServiceTemplateContext(BaseSystemdServiceTemplateContext):
    """
    Context values for the systemd service file content.
    """

    verbose_service_name: str = 'tinkerforge2mqtt'
    exec_start: str = f'{sys.argv[0]} publish-loop'


@dataclasses.dataclass
class SystemdServiceInfo(BaseSystemdServiceInfo):
    """
    Information for systemd helper functions.
    """

    template_context: SystemdServiceTemplateContext = dataclasses.field(default_factory=SystemdServiceTemplateContext)


@dataclasses.dataclass
class UserSettings:
    """
    Tinkerforge -> MQTT - settings

    Note: Insert at least device address + key and your MQTT settings.

    See README for more information.
    """

    host: str = 'localhost'
    port: int = 4223

    callback_period: int = 1  # in seconds

    # Information about the MQTT server:
    mqtt: dataclasses = dataclasses.field(default_factory=MqttSettings)

    systemd: dataclasses = dataclasses.field(default_factory=SystemdServiceInfo)



import abc
import logging

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice
from paho.mqtt.client import Client
from rich import print  # noqa
from tinkerforge.ip_connection import Device

from tinkerforge2mqtt.device_map_utils.generics import iter_interest_functions
from tinkerforge2mqtt.device_map_utils.led_config import BrickletLedConfigSelect
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator
from tinkerforge2mqtt.user_settings import UserSettings


logger = logging.getLogger(__name__)


class DeviceMapBase(abc.ABC):
    device_identifier: int

    @abc.abstractmethod
    def __init__(
        self,
        *,
        main_mqtt_device: MainMqttDevice,
        device: Device,
        mqtt_client: Client,
        user_settings: UserSettings,
    ):
        self.main_mqtt_device = main_mqtt_device
        self.device = device
        self.mqtt_client = mqtt_client
        self.user_settings = user_settings

        self.mqtt_device = MqttDevice(
            main_device=main_mqtt_device,
            name=f'{device.DEVICE_DISPLAY_NAME} ({device.uid_string})',
            uid=device.uid_string,
            manufacturer='Tinkerforge',
            model=device.DEVICE_DISPLAY_NAME,
            sw_version=self.get_sw_version(),
        )

        self.setup_sensors()
        self.setup_callbacks()
        self.poll()

    @abc.abstractmethod
    def setup_sensors(self):
        if hasattr(self.device, 'get_chip_temperature'):
            self.chip_temperature_sensor = Sensor(
                device=self.mqtt_device,
                name='Chip Temperature',
                uid='chip_temperature',
                device_class='temperature',
                state_class='measurement',
                unit_of_measurement='°C',
                suggested_display_precision=0,
            )
            logger.info(f'Sensor: {self.chip_temperature_sensor}')

        if hasattr(self.device, 'get_status_led_config'):
            self.led_config_sensor = BrickletLedConfigSelect(device=self.device, mqtt_device=self.mqtt_device)
            logger.info(f'Sensor: {self.led_config_sensor}')

    @abc.abstractmethod
    def setup_callbacks(self):
        pass

    def iter_known_functions(self, device: Device):
        assert (
            device.DEVICE_IDENTIFIER == self.device_identifier
        ), f'Wrong device: {device} is not {self.device_identifier}'

        yield from iter_interest_functions(device)

    @print_exception_decorator
    def poll(self):
        logger.info(f'Polling {self.device.DEVICE_DISPLAY_NAME} ({self.device.uid_string})')

        if get_chip_temperature := getattr(self.device, 'get_chip_temperature', None):
            value = get_chip_temperature()
            logger.debug(f'{self.device.DEVICE_DISPLAY_NAME} chip temperature: {value}°C')
            self.chip_temperature_sensor.set_state(state=value)
            self.chip_temperature_sensor.publish(self.mqtt_client)
            logger.info(f'Chip temperature: {value}°C: {self.chip_temperature_sensor}')

        if hasattr(self, 'led_config_sensor'):
            self.led_config_sensor.poll(self.mqtt_client)

        self.main_mqtt_device.poll_and_publish(self.mqtt_client)

    def get_sw_version(self) -> str:
        api_version = self.device.get_api_version()
        sw_version = '.'.join(str(number) for number in api_version)
        return sw_version

    def __str__(self):
        return f'{self.__class__.__name__} (UID: {self.device.uid_string})'

    def __repr__(self):
        return f'<{self}>'



import logging

from ha_services.mqtt4homeassistant.components import BaseComponent
from ha_services.mqtt4homeassistant.components.select import Select
from paho.mqtt.client import Client

from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator


logger = logging.getLogger(__name__)

CONFIG2LED_OPTIONS = {
    0: 'Off',
    1: 'On',
    2: 'Heartbeat',
    3: 'Status',
}
LED_OPTIONS2CONFIG = {v: k for k, v in CONFIG2LED_OPTIONS.items()}


class BrickletLedConfigSelect:
    def __init__(self, *, device, mqtt_device):
        self.device = device
        self.mqtt_device = mqtt_device

        current_config = self.device.get_status_led_config()

        self.led_config_select = Select(
            device=self.mqtt_device,
            name='LED Config',
            uid='led_config',
            callback=self.callback,
            options=tuple(CONFIG2LED_OPTIONS.values()),
            default_option=CONFIG2LED_OPTIONS[current_config],
        )

    @print_exception_decorator
    def poll(self, mqtt_client: Client):
        value = self.device.get_status_led_config()
        option = CONFIG2LED_OPTIONS[value]
        logger.info(f'{self.device.DEVICE_DISPLAY_NAME} status LED config: {value=} {option=}')
        self.led_config_select.set_state(state=option)
        self.led_config_select.publish(mqtt_client)

    def callback(self, *, client: Client, component: BaseComponent, old_state: str, new_state: str):
        logger.info(f'{component.name} state changed: {old_state!r} -> {new_state!r}')
        new_config = LED_OPTIONS2CONFIG[new_state]
        self.device.set_status_led_config(config=new_config)
        self.poll(mqtt_client=client)



import logging

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from tinkerforge.bricklet_temperature_v2 import BrickletTemperatureV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickletTemperatureV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricks/HATZero_Brick_Python.html

    device_identifier = BrickletTemperatureV2.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletTemperatureV2, **kwargs):
        self.device: BrickletTemperatureV2 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()
        self.temperature = Sensor(
            device=self.mqtt_device,
            name='Temperature',
            uid='temperature',
            device_class='temperature',
            state_class='measurement',
            unit_of_measurement='°C',
            suggested_display_precision=2,
        )

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.set_temperature_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option=self.device.THRESHOLD_OPTION_OFF,
            min=-999,
            max=999,
        )
        self.device.register_callback(self.device.CALLBACK_TEMPERATURE, self.callback_temperature)

    @print_exception_decorator
    def callback_temperature(self, value):
        temperature = value / 100
        logger.info(f'Temperature callback: {temperature}°C (UID: {self.device.uid_string})')
        self.temperature.set_state(temperature)
        self.temperature.publish(self.mqtt_client)



import logging
import time

import rich_click as click
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE
from ha_services.mqtt4homeassistant.mqtt import get_connected_client
from rich import print  # noqa
from rich import get_console
from tinkerforge.ip_connection import IPConnection

from tinkerforge2mqtt.cli_app import cli
from tinkerforge2mqtt.cli_app.settings import get_user_settings
from tinkerforge2mqtt.device_registry.devices_handler import DevicesHandler
from tinkerforge2mqtt.user_settings import UserSettings


logger = logging.getLogger(__name__)


def setup_logging(*, verbosity: int, log_format='%(message)s'):  # Move to cli_tools
    if verbosity == 0:
        level = logging.ERROR
    elif verbosity == 1:
        level = logging.WARNING
    elif verbosity == 2:
        level = logging.INFO
    else:
        level = logging.DEBUG
        if '%(name)s' not in log_format:
            log_format = f'(%(name)s) {log_format}'

    console = get_console()
    console.print(f'(Set log level {verbosity}: {logging.getLevelName(level)})', justify='right')
    logging.basicConfig(
        level=level,
        format=log_format,
        datefmt='[%x %X.%f]',
        # handlers=[
        #     RichHandler(console=console, omit_repeated_times=False,
        #     log_time_format='[%x FOO %X]'
        #
        # )],
        force=True,
    )


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE | {'default': 0})
def publish_loop(verbosity: int):
    """
    Publish Tinkerforge devices events via MQTT to Home Assistant.
    """
    setup_logging(verbosity=verbosity, log_format='%(levelname)s %(processName)s %(threadName)s %(message)s')
    user_settings: UserSettings = get_user_settings(verbosity=verbosity)

    # https://www.tinkerforge.com/en/doc/Software/IPConnection_Python.html
    ipcon = IPConnection()
    connect_kwargs = dict(
        host=user_settings.host,
        port=user_settings.port,
    )
    print(f'Connecting to {connect_kwargs}')
    ipcon.connect(**connect_kwargs)

    mqtt_client = get_connected_client(settings=user_settings.mqtt, verbosity=verbosity)
    mqtt_client.loop_start()

    devices_handler = DevicesHandler(ipcon, mqtt_client=mqtt_client, user_settings=user_settings)

    ipcon.register_callback(IPConnection.CALLBACK_ENUMERATE, devices_handler)

    while True:
        try:
            ipcon.enumerate()
            time.sleep(5)
        except KeyboardInterrupt:
            logger.info('Keyboard interrupt')
            ipcon.disconnect()
            break



import logging

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from tinkerforge.brick_hat_zero import BrickHATZero

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickHATZeroMapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricks/HATZero_Brick_Python.html

    device_identifier = BrickHATZero.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickHATZero, **kwargs):
        self.device: BrickHATZero = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()
        self.usb_voltage_sensor = Sensor(
            device=self.mqtt_device,
            name='USB Voltage',
            uid='voltage',
            device_class='voltage',
            state_class='measurement',
            unit_of_measurement='V',
            suggested_display_precision=3,
        )
        logger.info(f'Sensor: {self.usb_voltage_sensor}')

    @print_exception_decorator
    def setup_callbacks(self):
        super().setup_callbacks()
        self.device.set_usb_voltage_callback_configuration(
            period=self.user_settings.callback_period * 1000,
            value_has_to_change=False,
            option='x',  # Threshold is turned off
            min=0,
            max=999,
        )
        self.device.register_callback(BrickHATZero.CALLBACK_USB_VOLTAGE, self.callback_usb_voltage)
        self.callback_usb_voltage(value=self.device.get_usb_voltage())

    @print_exception_decorator
    def callback_usb_voltage(self, value):
        logger.debug(f'USB Voltage: {value / 1000}V (UID: {self.device.uid_string})')
        self.usb_voltage_sensor.set_state(state=value / 1000)
        self.usb_voltage_sensor.publish(self.mqtt_client)



import logging

from ha_services.mqtt4homeassistant.components.switch import Switch
from paho.mqtt.client import Client
from tinkerforge.bricklet_solid_state_relay_v2 import BrickletSolidStateRelayV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.device_map_utils.utils import print_exception_decorator

logger = logging.getLogger(__name__)


@register_map_class()
class BrickletSolidStateRelayV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/SolidStateRelayV2_Bricklet_Python.html

    device_identifier = BrickletSolidStateRelayV2.DEVICE_IDENTIFIER

    def __init__(self, *, device: BrickletSolidStateRelayV2, **kwargs):
        self.device: BrickletSolidStateRelayV2 = device
        super().__init__(device=device, **kwargs)

    @print_exception_decorator
    def setup_sensors(self):
        super().setup_sensors()

        self.relay_switch = Switch(
            device=self.mqtt_device,
            name='Relay',
            uid='relay',
            callback=self.relay_callback,
        )
        logger.info(f'Creating: {self.relay_switch}')

    @print_exception_decorator
    def setup_callbacks(self):
        logger.info(f'setup_callbacks {self}')
        super().setup_callbacks()

        self.device.register_callback(self.device.CALLBACK_MONOFLOP_DONE, self.callback_monoflop_done)

    @print_exception_decorator
    def callback_monoflop_done(self, value):
        logger.warning(f'TODO: Monoflop Done: {value} (UID: {self.device.uid_string})')

    @print_exception_decorator
    def poll(self):
        super().poll()

        state: bool = self.device.get_state()
        logger.info(f'Polling {state=} from {self.relay_switch}')
        if state:
            self.relay_switch.set_state(self.relay_switch.ON)
        else:
            self.relay_switch.set_state(self.relay_switch.OFF)
        self.relay_switch.publish(self.mqtt_client)

    @print_exception_decorator
    def relay_callback(self, *, client: Client, component: Switch, old_state: str, new_state: str):
        logger.error(f'{component.name} state changed: {old_state!r} -> {new_state!r}')
        if new_state == self.relay_switch.ON:
            self.device.set_state(True)
        else:
            self.device.set_state(False)

        self.poll()



import sys

import rich_click as click
from cli_base.cli_tools import git_history
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE, setup_logging
from rich import print  # noqa
from tinkerforge2mqtt.cli_app import cli


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def update_readme_history(verbosity: int):
    """
    Update project history base on git commits/tags in README.md

    Will be exited with 1 if the README.md was updated otherwise with 0.

    Also, callable via e.g.:
        python -m cli_base update-readme-history -v
    """
    setup_logging(verbosity=verbosity)
    updated = git_history.update_readme_history(verbosity=verbosity)
    exit_code = 1 if updated else 0
    if verbosity:
        print(f'{exit_code=}')
    sys.exit(exit_code)
